/* Sylpheed Pro full-text search plugin
 * Copyright (C) 2009 SRA OSS, Inc. Japan
 */

#include "config.h"

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gmodule.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef G_OS_WIN32
#  include <process.h>
#endif

#include "sylph/defs.h"
#include "sylph/sylmain.h"
#include "sylph/folder.h"
#include "sylph/procmsg.h"
#include "sylph/utils.h"
#include "plugin.h"

#include "dbimport.h"
#include "common.h"
#include "sylph-searcher.h"

#include "icons/sylpheed-pro-icon-32.h"

static SylPluginInfo info = {
	"Sylpheed Pro Full-text Search Plug-in",
	VERSION,
	"Coypright (C) 2009 SRA OSS, Inc. Japan",
	"Full-text search plug-in of Sylpheed Pro"
};

static struct _proc
{
	pid_t pid;
	gint fds1[2];
	gint fds2[2];
	FILE *rfp;
	FILE *wfp;
} ext_proc;

GtkWidget *status_icon = NULL;

static gboolean plugin_init_success = FALSE;
static gboolean auth_success = FALSE;
static gint auth_state = 0;
static gboolean setup_success = FALSE;

gboolean ext_send_msg(const gchar *format, ...)
{
	va_list args;
	gchar buf[BUFFSIZE];
	size_t len;

	if (!ext_proc.wfp)
		return FALSE;

	va_start(args, format);
	g_vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	len = strlen(buf);
	if (len == sizeof(buf) - 1)
		len--;
	buf[len] = '\n';
	buf[len + 1] = '\0';

	//g_print("dbimport: send: %s", buf);
	if (fputs(buf, ext_proc.wfp) == EOF) {
		g_warning("dbimport: ext_send_msg: fputs failed");
		return FALSE;
	}
	fflush(ext_proc.wfp);

	return TRUE;
}

gboolean ext_recv_msg(gchar *buf, size_t len)
{
	if (!ext_proc.rfp)
		return FALSE;

	if (fgets(buf, len, ext_proc.rfp) == NULL) {
		g_warning("dbimport: ext_recv_msg: fgets failed");
		return FALSE;
	}

	g_strchomp(buf);
	//g_print("dbimport: read: %s\n", buf);
	return TRUE;
}

gboolean ext_recv_line(gchar **line)
{
	gchar buf[BUFFSIZE];
	gchar *str = NULL;
	gint len = 0;

	while (1) {
		if (fgets(buf, sizeof(buf), ext_proc.rfp) == NULL) {
			g_warning("dbimport: ext_recv_line: fgets failed");
			return FALSE;
		}
		if (!str) {
			len = strlen(buf);
			str = g_strdup(buf);
		} else {
			gint prev_len = len;
			len += strlen(buf);
			str = g_realloc(str, len + 1);
			strcpy(str + prev_len, buf);
		}
		if (len > 0 && str[len - 1] == '\n')
			break;
	}

	*line = str;
	return TRUE;
}

gboolean ext_recv_bytes(guint bytes, gchar **str)
{
	gchar *data;
	gchar *p;
	size_t len, left;

	data = g_malloc(bytes + 1);
	if (!data)
		return FALSE;
	p = data;
	left = bytes;

	while (left > 0) {
		len = MIN(BUFFSIZE, left);
		if (fread(p, len, 1, ext_proc.rfp) != 1) {
			g_free(data);
			return FALSE;
		}
		left -= len;
		p += len;
	}
	*p = '\0';

	*str = data;
	return TRUE;
}

static gint check_license_key_parse_retval(const gchar *val)
{
	if (g_ascii_strcasecmp(val, "OK") == 0) {
		return 0;
	} else if (g_ascii_strncasecmp(val, "TRIAL", 5) == 0) {
		const gchar *p = val + 5;
		gint day_left;

		while (g_ascii_isspace(*p))
			p++;
		day_left = atoi(p);
		return 1 + day_left;
	} else if (g_ascii_strcasecmp(val, "EXPIRED") == 0) {
		return -2;
	} else if (g_ascii_strncasecmp(val, "ERROR", 5) == 0) {
		return -1;
	} else
		return -128;
}

gint check_license_key(const gchar *user, const gchar *key)
{
	gchar buf[64] = "";

	ext_send_msg("LICENSEKEY\n%s\n%s", user, key);
	ext_recv_msg(buf, sizeof(buf));
	return check_license_key_parse_retval(buf);
}

gint check_license_key_check_only(const gchar *user, const gchar *key)
{
	gchar buf[64] = "";

	ext_send_msg("CHECKLICENSEKEY\n%s\n%s", user, key);
	ext_recv_msg(buf, sizeof(buf));
	return check_license_key_parse_retval(buf);
}

gint get_trial_key(gchar *buf, size_t len)
{
	*buf = '\0';
	ext_send_msg("GETTRIALKEY");
	ext_recv_msg(buf, len);

	if (*buf == '\0' || g_ascii_strncasecmp(buf, "ERROR", 5) == 0) {
		return -1;
	} else {
		return 0;
	}
}

gint check_auth_status(void)
{
	gchar buf[64] = "";

	ext_send_msg("AUTHSTATUS");
	ext_recv_msg(buf, sizeof(buf));
	debug_print("auth status: %s\n", buf);
	if (g_ascii_strcasecmp(buf, "OK") == 0) {
		auth_success = TRUE;
		auth_state = 0;
	} else if (g_ascii_strcasecmp(buf, "TRIAL") == 0) {
		auth_success = TRUE;
		auth_state = 1;
	} else if (g_ascii_strcasecmp(buf, "EXPIRED") == 0) {
		debug_print("trial key expired\n");
		auth_state = -2;
	} else {
		debug_print("auth failed\n");
		auth_state = -1;
	}

	return auth_state;
}

static void update_menu_state(void)
{
	gchar buf[BUFFSIZE];
	AppConfig *config;

	config = searcher_get_config();

	g_snprintf(buf, sizeof(buf),
		   "<Main>%s", _("/Edit/Full-text search (Sylpheed Pro)"));
	if (config->enable_searcher && auth_success && setup_success) {
		syl_plugin_menu_set_sensitive(buf, TRUE);
		if (status_icon)
			gtk_widget_show(status_icon);
	} else {
		syl_plugin_menu_set_sensitive(buf, FALSE);
		if (status_icon)
			gtk_widget_hide(status_icon);
	}
}

gint start_database(void)
{
	AppConfig *config;
	gint ret = -1;

	config = searcher_get_config();

	if (config->enable_searcher) {
		check_auth_status();
		if (auth_success) {
			gchar buf[64] = "";

			setup_success = TRUE;
			debug_print("starting database...\n");
			ext_send_msg("STARTDB");
			ext_recv_msg(buf, sizeof(buf));
			if (g_ascii_strcasecmp(buf, "OK") == 0) {
				debug_print("database is running\n");
				ret = 0;
				buf[0] = '\0';
				ext_send_msg("CHECKDB");
				ext_recv_msg(buf, sizeof(buf));
				if (g_ascii_strcasecmp(buf, "OK") != 0) {
					g_warning("checkdb failed");
					ret = -1;
				}
			} else {
				debug_print("startup of database failed\n");
			}
		}

		update_menu_state();
	}

	return ret;
}

static void init_done_cb(GObject *obj, gpointer data)
{
	gchar buf[64] = "";
	AppConfig *config;
	gboolean db_ok = FALSE;

	debug_print("dbimport: (%p) init_done_cb\n", obj);

	config = searcher_get_config();

	if (searcher_read_config(FALSE) < 0) {
		if (searcher_setup() == 0) {
			setup_success = TRUE;
			db_ok = TRUE;
		}
	} else if (config->enable_searcher) {
		gchar *key;
		gint state;

		if ((key = read_license_key()) != NULL) {
			state = check_license_key("user", key);
			if (state >= 0) {
				setup_success = TRUE;
				debug_print("starting database...\n");
				ext_send_msg("STARTDB");
				ext_recv_msg(buf, sizeof(buf));
				if (g_ascii_strcasecmp(buf, "OK") == 0) {
					buf[0] = '\0';
					ext_send_msg("CHECKDB");
					ext_recv_msg(buf, sizeof(buf));
					if (g_ascii_strcasecmp(buf, "OK") != 0) {
						g_warning("checkdb failed");
					} else
						db_ok = TRUE;
				} else {
					debug_print("startup of database failed\n");
				}

				if (state >= 1 && state <= 2) {
					gchar msg[256];

					debug_print("license key will expire soon: %s (left: %d)\n", key, state - 1);
					g_snprintf(msg, sizeof(msg), _("Trial version of Sylpheed Pro will expire soon\n(%d day(s) left)."), state - 1);
					alert_dialog(_("Sylpheed Pro trial will expire soon"), msg,
						     GTK_WINDOW(syl_plugin_manage_window_get_focus_window()));
				}
			} else {
				g_warning("license key check failed: %s (%d)",
					  key, state);
				if (state == -2) {
					g_warning("trial version expired");
					alert_dialog(_("Sylpheed Pro trial expired"),
						     _("Trial verson of Sylpheed Pro has been expired."),
						     GTK_WINDOW(syl_plugin_manage_window_get_focus_window()));
				}
			}
			g_free(key);
		} else {
			g_warning("reading license key file failed");
		}
	}

	if (config->enable_searcher) {
		check_auth_status();
	}
	update_menu_state();
}

static void app_exit_cb(GObject *obj, gpointer data)
{
	debug_print("dbimport: (%p) app_exit_cb\n", obj);
}

static void folderlist_updated_cb(GObject *obj, gpointer data)
{
	gchar buf[BUFFSIZE];

	debug_print("dbimport: (%p) folderlist_updated_cb\n", obj);

	if (!auth_success || !setup_success)
		return;

	ext_send_msg("RELOAD");
	ext_recv_msg(buf, sizeof(buf));
}

static gboolean item_is_excluded(const gchar *id, AppConfig *config)
{
	const gchar *p = id;
	gchar **exclv;
	gint i;

	if (*p == '#') {
		p++;
		while (g_ascii_isalpha(*p)) p++;
	}
	while (*p == '/') p++;

	exclv = config->exclude_list_v;
	if (!exclv)
		return FALSE;

	for (i = 0; exclv[i] != NULL; i++) {
		if (exclv[i][0] != '\0' && strcasestr(p, exclv[i])) {
			debug_print("exclude: %s in %s\n", exclv[i], p);
			return TRUE; 
		}
	}

	return FALSE;
}

static void add_msg_cb(GObject *obj, FolderItem *item, const gchar *file,
		       guint num, gpointer data)
{
	gchar *id;
	gchar buf[BUFFSIZE];
	AppConfig *config;

	g_return_if_fail(item != NULL);

	if (!auth_success || !setup_success)
		return;

	config = searcher_get_config();
	if (!config->enable_searcher || !config->auto_update)
		return;

	if (item->stype == F_QUEUE || item->stype == F_TRASH) {
		debug_print("dbimport: skipping queue/trash msg: %s\n", file);
		return;
	}

	id = folder_item_get_identifier(item);
	g_return_if_fail(id != NULL);

	if (item_is_excluded(id, config)) {
		g_free(id);
		return;
	}

	//g_print("dbimport: (%p) add_msg_cb: %s: %s %u\n", obj, id, file, num);
	ext_send_msg("ADDMSG\n%s\n%u\n%s", id, num, file);
	ext_recv_msg(buf, sizeof(buf));

	g_free(id);
}

static void remove_msg_cb(GObject *obj, FolderItem *item, const gchar *file,
			  guint num, gpointer data)
{
	gchar *id;
	gchar buf[BUFFSIZE];
	AppConfig *config;

	if (!auth_success || !setup_success)
		return;

	config = searcher_get_config();
	if (!config->enable_searcher || !config->auto_update)
		return;

	id = folder_item_get_identifier(item);
	g_return_if_fail(id != NULL);

	debug_print("dbimport: (%p) remove_msg_cb: %s: %s\n", obj, id, file);
	ext_send_msg("REMOVEMSG\n%s\n%u\n%s", id, num, file);
	ext_recv_msg(buf, sizeof(buf));

	g_free(id);
}

static void remove_all_msg_cb(GObject *obj, FolderItem *item)
{
	gchar *id;
	gchar buf[BUFFSIZE];
	AppConfig *config;

	if (!auth_success || !setup_success)
		return;

	config = searcher_get_config();
	if (!config->enable_searcher || !config->auto_update)
		return;

	id = folder_item_get_identifier(item);
	g_return_if_fail(id != NULL);

	debug_print("dbimport: (%p) remove_all_msg_cb: %s\n", obj, id);
	ext_send_msg("REMOVEALLMSG\n%s", id);
	ext_recv_msg(buf, sizeof(buf));

	g_free(id);
}

static void remove_folder_cb(GObject *obj, FolderItem *item)
{
	gchar *id;
	gchar buf[BUFFSIZE];
	AppConfig *config;

	if (!auth_success || !setup_success)
		return;

	config = searcher_get_config();
	if (!config->enable_searcher || !config->auto_update)
		return;

	id = folder_item_get_identifier(item);
	g_return_if_fail(id != NULL);

	debug_print("dbimport: (%p) remove_folder_cb: %s\n", obj, id);
	ext_send_msg("REMOVEFOLDER\n%s", id);
	ext_recv_msg(buf, sizeof(buf));

	g_free(id);
}

static void move_folder_cb(GObject *obj, FolderItem *item, const gchar *old_id,
			   const gchar *new_id)
{
	gchar buf[BUFFSIZE];
	AppConfig *config;

	if (!auth_success || !setup_success)
		return;

	config = searcher_get_config();
	if (!config->enable_searcher || !config->auto_update)
		return;

	debug_print("dbimport: move_folder_cb: %s -> %s\n", old_id, new_id);
	ext_send_msg("MOVEFOLDER\n%s\n%s", old_id, new_id);
	ext_recv_msg(buf, sizeof(buf));
}

typedef struct _ProgressDialog
{
	GtkWidget *dialog;
	GtkWidget *label;
	GtkWidget *progress;
	gboolean cancelled;
} ProgressDialog;

static void cancel_clicked(GtkWidget *widget, gpointer data)
{
	ProgressDialog *dialog = (ProgressDialog *)data;
	dialog->cancelled = TRUE;
}

static gint delete_cb(GtkWidget *widget, GdkEventAny *event, gpointer data)
{
	ProgressDialog *dialog = (ProgressDialog *)data;
	dialog->cancelled = TRUE;
	return TRUE;
}

static ProgressDialog *update_progress_dialog_create()
{
	GtkWidget *dialog;
	GtkWidget *label;
	GtkWidget *progress;
	GtkWidget *cancel_btn;
	ProgressDialog *prog_dialog;

	prog_dialog = g_new0(ProgressDialog, 1);

	dialog = gtk_dialog_new();
	gtk_widget_set_size_request(dialog, 480, -1);
	gtk_container_set_border_width(GTK_CONTAINER(dialog), 8);
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
	gtk_window_set_title(GTK_WINDOW(dialog), _("Updating search index"));
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
	gtk_widget_realize(dialog);
	syl_plugin_manage_window_set_transient(GTK_WINDOW(dialog));

	gtk_container_set_border_width
		(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 0);
	gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 8);
	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);

	label = gtk_label_new(_("Updating search index ..."));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label,
			   FALSE, FALSE, 0);
	gtk_widget_show(label);

	progress = gtk_progress_bar_new();
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), progress,
			   FALSE, FALSE, 0);
	gtk_progress_set_show_text(GTK_PROGRESS(progress), TRUE);
	gtk_progress_set_value(GTK_PROGRESS(progress), 0.0);
	gtk_progress_set_percentage(GTK_PROGRESS(progress), 0.0);
	gtk_widget_show(progress);

	cancel_btn = gtk_dialog_add_button(GTK_DIALOG(dialog),
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_NONE);
	gtk_widget_grab_default(cancel_btn);
	gtk_widget_grab_focus(cancel_btn);

	prog_dialog->dialog = dialog;
	prog_dialog->label = label;
	prog_dialog->progress = progress;
	prog_dialog->cancelled = FALSE;

	g_signal_connect(cancel_btn, "clicked", G_CALLBACK(cancel_clicked),
			 prog_dialog);
	g_signal_connect(dialog, "delete-event", G_CALLBACK(delete_cb),
			 prog_dialog);

	gtk_widget_show_now(dialog);

	return prog_dialog;
}

gboolean confirm_dialog(const gchar *title, const gchar *msg,
			const gchar *check_msg, gboolean *check_active)
{
	GtkWidget *dialog;
	GtkWidget *checkbtn;
	gint response;
	gboolean active;
	GtkWindow *parent;

	parent = GTK_WINDOW(syl_plugin_manage_window_get_focus_window());

	dialog = gtk_message_dialog_new(parent,
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_QUESTION,
					GTK_BUTTONS_YES_NO,
					"%s", title);
#if GTK_CHECK_VERSION(2, 6, 0)
	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
						 "%s", msg);
#endif
	checkbtn = gtk_check_button_new_with_label(check_msg);
	gtk_widget_show(checkbtn);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), checkbtn,
			   FALSE, FALSE, 0);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_YES);

	gtk_widget_show(dialog);
	gtk_window_set_transient_for(GTK_WINDOW(dialog), parent);
	response = gtk_dialog_run(GTK_DIALOG(dialog));

	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbtn));

	gtk_widget_destroy(dialog);

	*check_active = active;

	debug_print("response: %d, active: %d\n", response, active);
	return (response == GTK_RESPONSE_YES);
}

static gint update_folder_item(ProgressDialog *dialog, FolderItem *item,
			       gboolean recursive)
{
	gchar *id;
	gint ret = 0;

	g_return_val_if_fail(item != NULL, -1);

	if (item->stype == F_VIRTUAL || item->stype == F_QUEUE ||
	    item->stype == F_TRASH) {
		return 0;
	}

	if (item->path) {
		gchar buf[BUFFSIZE];
		gint total, cur = 0;
		GTimeVal tv_prev, tv_cur;

		id = folder_item_get_identifier(item);
		if (!id)
			return -1;

		if (item_is_excluded(id, searcher_get_config())) {
			g_free(id);
			return 0;
		}

		g_snprintf(buf, sizeof(buf), _("Updating %s ..."), id);
		debug_print("%s\n", buf);
		gtk_label_set_text(GTK_LABEL(dialog->label), buf);
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR(dialog->progress), "");
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dialog->progress), 0.0);
		g_get_current_time(&tv_prev);
		ui_update();

		ext_send_msg("ADDFOLDER\n%s", id);
		if (!ext_recv_msg(buf, sizeof(buf))) {
			g_free(id);
			return -1;
		}
		debug_print("dbimport: recv: %s\n", buf);
		if (sscanf(buf, "TOTAL %d", &total) != 1) {
			g_warning("error: %s", buf);
			g_free(id);
			return -1;
		}

		while (total > 0 && cur < total) {
			ext_recv_msg(buf, sizeof(buf));
			if (sscanf(buf, "ADD %d", &cur) != 1) {
				ret = -1;
				g_warning("error: %s", buf);
				break;
			}
			g_get_current_time(&tv_cur);
			if ((tv_cur.tv_sec - tv_prev.tv_sec) * G_USEC_PER_SEC +
			    tv_cur.tv_usec - tv_prev.tv_usec > 100 * 1000) {
				g_snprintf(buf, sizeof(buf), "%d / %d", cur, total);
				gtk_progress_bar_set_text(GTK_PROGRESS_BAR(dialog->progress), buf);
				gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dialog->progress), (gfloat)cur / (gfloat)total);
				ui_update();
				tv_prev = tv_cur;
			}
			if (cur < total) {
				if (dialog->cancelled) {
					debug_print("update cancelled\n");
					ext_send_msg("CANCEL");
					ret = -2;
					break;
				}
				ext_send_msg("NEXT");
			}
		}

		g_snprintf(buf, sizeof(buf), "%d / %d", cur, total);
		gtk_progress_bar_set_text(GTK_PROGRESS_BAR(dialog->progress), buf);
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(dialog->progress), (gfloat)cur / (gfloat)total);
		ui_update();

		if (ret != -1) {
			ext_recv_msg(buf, sizeof(buf));
			debug_print("dbimport: recv: %s\n", buf);
		}

		// cleanup removed
		if (ret == 0) {
			debug_print("removing nonexist messages in %s from DB\n", id);
			ext_send_msg("REMOVENOTEXIST\n%s", id);
			ext_recv_msg(buf, sizeof(buf));
		}

		debug_print("done.\n");
		g_free(id);
	}

	if (ret == 0 && recursive && item->node && item->node->children) {
		GNode *child = item->node->children;

		while (child) {
			ret = update_folder_item(dialog,
						 FOLDER_ITEM(child->data),
						 recursive);
			if (ret < 0)
				return ret;
			child = child->next;
		}
	}

	return ret;
}

static void update_folder_index(void)
{
	FolderItem *item;
	gint response;
	gboolean recursive;
	ProgressDialog *dialog;
	AppConfig *config;

	config = searcher_get_config();
	if (!config->enable_searcher)
		return;

	item = syl_plugin_folderview_get_selected_item();
	if (!item) {
		g_warning("update_folder_index: item not found.");
		return;
	}

	syl_plugin_inc_lock();

	response = confirm_dialog(_("Update folder index"),
				  _("Do you want to update the folder's search index?\n(it may take some time)"),
				  _("Update subfolders"), &recursive);

	if (response) {
		dialog = update_progress_dialog_create();
		if (update_folder_item(dialog, item, recursive) < 0 &&
		    !dialog->cancelled) {
			alert_dialog(_("Error"),
				     _("Error occurred while updating folder index."),
				     GTK_WINDOW(dialog->dialog));
		}
		gtk_widget_destroy(dialog->dialog);
		g_free(dialog);
	}

	syl_plugin_inc_unlock();
}

static void searcher_window_open(void)
{
	searcher_init();
}

void searcher_config_open(void)
{
	searcher_settings();
	if (searcher_read_config(FALSE) == 0) {
		setup_success = TRUE;
		check_auth_status();
	}
	update_menu_state();
}

static void searcher_window_open_with_folder(void)
{
	FolderItem *item;
	gchar *id;

	item = syl_plugin_folderview_get_selected_item();
	id = folder_item_get_identifier(item);

	searcher_init();
	searcher_clear();
	searcher_set_folder(id);

	g_free(id);
}

static GtkWidget *about_window;

static void response_cb(GtkDialog *dialog, gint response_id, gpointer data)
{
	gtk_widget_destroy(GTK_WIDGET(dialog));
	about_window = NULL;
}

void about_dialog(void)
{
        GtkWidget *dialog;
        GtkWidget *vbox;
        GtkWidget *label;
	gchar buf[256];

        if (about_window) {
                gtk_window_present(GTK_WINDOW(about_window));
                return;
        }

        dialog = gtk_dialog_new_with_buttons(_("About"), NULL, 0,
                                             GTK_STOCK_OK, GTK_RESPONSE_OK,
                                             NULL);
        gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
        gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
        gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

        vbox = gtk_vbox_new(FALSE, 4);
        gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
        gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
                           FALSE, FALSE, 0);

        label = gtk_label_new(info.name);
        gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	g_snprintf(buf, sizeof(buf), "version %s", info.version);
        label = gtk_label_new(buf);
        gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
        label = gtk_label_new(info.author);
        gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
        label = gtk_label_new(info.description);
        gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

        g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(response_cb),
                         NULL);

        gtk_widget_show_all(dialog);
        about_window = dialog;
}

static void menu_popup_func(GObject *obj, GtkItemFactory *ifactory,
			    gpointer data)
{
	FolderItem *item;
	GtkWidget *widget;
	AppConfig *config;

	config = searcher_get_config();
	if (!config->enable_searcher || !auth_success || !setup_success)
		return;

	item = syl_plugin_folderview_get_selected_item();
	if (!item)
		return;

	if (item->stype == F_VIRTUAL || item->stype == F_QUEUE ||
	    item->stype == F_TRASH)
		return;
	if (!FOLDER_IS_LOCAL(item->folder))
		return;

	widget = gtk_item_factory_get_item(ifactory, _("/Update search index (Sylpheed Pro)"));
	if (widget)
		gtk_widget_set_sensitive(widget, TRUE);
	widget = gtk_item_factory_get_item(ifactory, _("/Full-text search this folder (Sylpheed Pro)"));
	if (widget)
		gtk_widget_set_sensitive(widget, TRUE);
}

static gint exec_external_process(void)
{
	gchar fdrstr[20], fdwstr[20], pidstr[20];

	if (ext_proc.rfp) {
		debug_print("external process is already running.\n");
		return 0;
	}

	/* dbimport -> [fds1] -> dbimport-ext
	            <- [fds2] <-
	 */
	if (pipe(ext_proc.fds1) < 0) {
		g_warning("pipe failed");
		return -1;
	}
	if (pipe(ext_proc.fds2) < 0) {
		g_warning("pipe failed");
		close(ext_proc.fds1[0]);
		close(ext_proc.fds1[1]);
		return -1;
	}
	debug_print("pipe(parent): r %d w %d\n", ext_proc.fds2[0], ext_proc.fds1[1]);
	debug_print("pipe(child) : r %d w %d\n", ext_proc.fds1[0], ext_proc.fds2[1]);
	itos_buf(fdrstr, ext_proc.fds1[0]);
	itos_buf(fdwstr, ext_proc.fds2[1]);
	g_snprintf(pidstr, sizeof(pidstr), "%d", getpid());
#ifdef G_OS_WIN32
	if ((ext_proc.pid = _spawnlp(_P_NOWAIT, "dbimport-ext", "dbimport-ext", fdrstr, fdwstr, pidstr, NULL)) < 0) {
		perror("_spawnlp");
		close(ext_proc.fds1[0]);
		close(ext_proc.fds1[1]);
		close(ext_proc.fds2[0]);
		close(ext_proc.fds2[1]);
		return -1;
	}
#else
	if ((ext_proc.pid = fork()) == 0) {
		if (execlp("dbimport-ext", "dbimport-ext", fdrstr, fdwstr, pidstr, NULL) < 0) {
			perror("execl");
			_exit(1);
		}
	} else if (ext_proc.pid < 0) {
		perror("fork");
		close(ext_proc.fds1[0]);
		close(ext_proc.fds1[1]);
		close(ext_proc.fds2[0]);
		close(ext_proc.fds2[1]);
		return -1;
	}
#endif
	debug_print("executed dbimport-ext (pid: %d)\n", ext_proc.pid);
	close(ext_proc.fds1[0]);
	close(ext_proc.fds2[1]);

	ext_proc.rfp = fdopen(ext_proc.fds2[0], "rb");
	ext_proc.wfp = fdopen(ext_proc.fds1[1], "wb");

	return 0;
}

gint start_plugin_system(void)
{
	GdkPixbuf *pixbuf;
	GtkWidget *statusbar;

	if (plugin_init_success) {
		debug_print("start_plugin_system: already started.\n");
		return 0;
	}

	if (exec_external_process() < 0) {
		g_warning("start_plugin_system: exec_external_process failed.");
		return -1;
	}

	g_signal_connect(syl_app_get(), "add-msg", G_CALLBACK(add_msg_cb),
			 NULL);
	g_signal_connect(syl_app_get(), "remove-msg", G_CALLBACK(remove_msg_cb),
			 NULL);
	g_signal_connect(syl_app_get(), "remove-all-msg",
			 G_CALLBACK(remove_all_msg_cb), NULL);
	g_signal_connect(syl_app_get(), "remove-folder",
			 G_CALLBACK(remove_folder_cb), NULL);
	g_signal_connect(syl_app_get(), "move-folder",
			 G_CALLBACK(move_folder_cb), NULL);

	statusbar = syl_plugin_main_window_get_statusbar();
	if (statusbar) {
		pixbuf = gdk_pixbuf_new_from_inline
			(sizeof(sylpheed_pro_icon_32), sylpheed_pro_icon_32,
			 FALSE, NULL);
		status_icon = gtk_image_new_from_pixbuf(pixbuf);
		g_object_unref(pixbuf);
		gtk_box_pack_start(GTK_BOX(statusbar), status_icon,
				   FALSE, FALSE, 0);
		gtk_box_reorder_child(GTK_BOX(statusbar), status_icon, 2);
		gtk_widget_show(status_icon);
	}

	debug_print("start_plugin_system: done.\n");

	plugin_init_success = TRUE;

	return 0;
}

void plugin_load(void)
{
	debug_print("dbimport plug-in loaded.\n");

	syl_init_gettext(GETTEXT_PACKAGE, LOCALEDIR);

	g_signal_connect(syl_app_get(), "init-done", G_CALLBACK(init_done_cb),
			 NULL);
	g_signal_connect(syl_app_get(), "app-exit", G_CALLBACK(app_exit_cb),
			 NULL);
	g_signal_connect(syl_app_get(), "folderlist-updated",
			 G_CALLBACK(folderlist_updated_cb), NULL);

	syl_plugin_add_factory_item("<Main>", "/Edit/---", NULL, NULL);
	syl_plugin_add_factory_item("<Main>", _("/Edit/Full-text search (Sylpheed Pro)"),
				searcher_window_open, NULL);
	syl_plugin_add_factory_item("<Main>", "/Configuration/---", NULL, NULL);
	syl_plugin_add_factory_item("<Main>", _("/Configuration/Full-text search settings (Sylpheed Pro)"), searcher_config_open, NULL);
	syl_plugin_add_factory_item("<MailFolder>", "/---", NULL, NULL);
	syl_plugin_add_factory_item("<MailFolder>", _("/Update search index (Sylpheed Pro)"),
				update_folder_index, NULL);
	syl_plugin_add_factory_item("<MailFolder>", _("/Full-text search this folder (Sylpheed Pro)"),
				searcher_window_open_with_folder, NULL);

	syl_plugin_signal_connect("folderview-menu-popup",
				  G_CALLBACK(menu_popup_func), NULL);

	if (searcher_read_config(FALSE) == 0) {
		AppConfig *config;
		config = searcher_get_config();
		if (!config->enable_searcher)
			return;
	}

	start_plugin_system();
}

void plugin_unload(void)
{
	ext_send_msg("QUIT");
	if (ext_proc.rfp) {
		fclose(ext_proc.rfp);
		ext_proc.rfp = NULL;
	}
	if (ext_proc.wfp) {
		fclose(ext_proc.wfp);
		ext_proc.wfp = NULL;
	}

	if (status_icon) {
		gtk_widget_destroy(status_icon);
		status_icon = NULL;
	}

	g_signal_handlers_disconnect_by_func(syl_app_get(), init_done_cb, NULL);
	g_signal_handlers_disconnect_by_func(syl_app_get(), app_exit_cb, NULL);
	g_signal_handlers_disconnect_by_func(syl_app_get(), folderlist_updated_cb, NULL);
	syl_plugin_signal_disconnect(menu_popup_func, NULL);

	if (plugin_init_success) {
		g_signal_handlers_disconnect_by_func(syl_app_get(), add_msg_cb, NULL);
		g_signal_handlers_disconnect_by_func(syl_app_get(), remove_msg_cb, NULL);
		g_signal_handlers_disconnect_by_func(syl_app_get(), remove_all_msg_cb, NULL);
		g_signal_handlers_disconnect_by_func(syl_app_get(), remove_folder_cb, NULL);
		g_signal_handlers_disconnect_by_func(syl_app_get(), move_folder_cb, NULL);
		plugin_init_success = FALSE;
	}

	debug_print("dbimport plug-in unloaded!\n");
}

SylPluginInfo *plugin_info(void)
{
	return &info;
}

gint plugin_interface_version(void)
{
	return SYL_PLUGIN_INTERFACE_VERSION;
}

