/* Sylpheed Pro full-text search plugin
 * Copyright (C) 2007-2009 Sylpheed Development Team
 * Copyright (C) 2009 SRA OSS, Inc. Japan
 */

#include "config.h"

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <time.h>

#include <sylph/sylmain.h>
#include <sylph/prefs_common.h>
#include <sylph/account.h>
#include <sylph/folder.h>
#include <sylph/procmsg.h>
#include <sylph/codeconv.h>
#include "plugin.h"

#if HAVE_LOCALE_H
#  include <locale.h>
#endif

#ifdef G_OS_WIN32
#  include <windows.h>
#  include <winreg.h>
#  include <wchar.h>
#  include <shlobj.h>
#  include <fcntl.h>
#endif

#include "common.h"
#include "dbimport.h"
#include "sylph-searcher.h"
#include "searcher-config.h"

#include "icons/sylpheed-pro-icon-32.h"
#include "icons/sylpheed-pro-logo-42.h"

enum
{
	COL_SUBJECT,
	COL_FROM,
	COL_DATE,
	COL_FOLDER,
	COL_MSGNUM,
	COL_MSGSID,
	N_COLS
};

typedef struct _QueryInfo {
	gchar *body;
	gchar *from;
	gchar *to;
	gchar *subject;
	gchar *date_gt;
	gchar *date_lt;
	gchar *folder;
	gboolean folder_with_sub;
	gboolean has_attach;

	gint res_limit;

	gchar *sep_body;
} QueryInfo;

static AppConfig config;

QueryInfo *query_info;

static struct SylFTSearchWindow {
	GtkWidget *window;

	GtkWidget *body_entry;

	GtkWidget *from_entry;
	GtkWidget *to_entry;
	GtkWidget *subject_entry;
	GtkWidget *dategt_entry;
	GtkWidget *datelt_entry;
	GtkWidget *folder_entry;
	GtkWidget *folder_btn;
	GtkWidget *folder_with_sub_chkbtn;
	GtkWidget *has_attach_chkbtn;

	GtkWidget *treeview;
	GtkListStore *store;

	GtkWidget *textview;
	GtkTextTag *tag;

	GtkWidget *status_label;

	GtkWidget *search_btn;
	GtkWidget *clear_btn;
	GtkWidget *close_btn;
} ftsearch_window;

const gchar *ui_def =
	"<ui>"
	"  <menubar name='MainMenu'>"
	"    <menu name='File' action='FileAction'>"
	"      <menuitem name='Close' action='CloseAction'/>"
	"    </menu>"
	"    <menu name='Edit' action='EditAction'>"
	"      <menuitem name='Copy' action='CopyAction'/>"
	"      <menuitem name='Paste' action='PasteAction'/>"
	"    </menu>"
	"    <menu name='Tools' action='ToolsAction'>"
	"      <menuitem name='Settings' action='SettingsAction'/>"
	"    </menu>"
	"    <menu name='Help' action='HelpAction'>"
	"      <menuitem name='About' action='AboutAction'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";

static gint do_query(const QueryInfo *qinfo);
static gchar *get_body_text(gulong sid);

static void update_text(const gchar *text);

static QueryInfo *query_info_new(const gchar *body, const gchar *from,
				 const gchar *to, const gchar *subject);
static void query_info_free(QueryInfo *qinfo);

static const gchar *errcode_to_str(gint error);

static void row_activated	(GtkTreeView		*treeview,
				 GtkTreePath		*path,
				 GtkTreeViewColumn	*column,
				 gpointer		 data);
static void selection_changed	(GtkTreeSelection	*selection,
				 gpointer		 data);

static void search_entry_activated	(GtkWidget	*widget,
					 gpointer	 data);

static gint search_deleted	(GtkWidget	*widget,
				 GdkEventAny	*event,
				 gpointer	 data);

static void folder_btn_clicked	(GtkButton	*button,
				 gpointer	 data);
static void clear_clicked	(GtkButton	*button,
				 gpointer	 data);
static void search_clicked	(GtkButton	*button,
				 gpointer	 data);
static void close_clicked	(GtkButton	*button,
				 gpointer	 data);

static void close_cb	(void);
static void copy_cb	(void);
static void paste_cb	(void);
static void settings_cb	(void);
static void about_cb	(void);

static GtkActionEntry action_entries[] = {
	{"FileAction", NULL, N_("_File"), NULL, NULL, NULL},
	{"CloseAction", NULL, N_("_Close"), NULL, NULL, close_cb},
	{"EditAction", NULL, N_("_Edit"), NULL, NULL, NULL},
	{"CopyAction", NULL, N_("_Copy"), NULL, NULL, copy_cb},
	{"PasteAction", NULL, N_("_Paste"), NULL, NULL, paste_cb},
	{"ToolsAction", NULL, N_("_Tools"), NULL, NULL, NULL},
	{"SettingsAction", NULL, N_("_Settings"), NULL, NULL, settings_cb},
	{"HelpAction", NULL, N_("_Help"), NULL, NULL, NULL},
	{"AboutAction", NULL, N_("_About"), NULL, NULL, about_cb},
};


static gint do_query(const QueryInfo *qinfo)
{
	gboolean have_cond = FALSE;
	gulong sid;
	gchar *from, *to, *subject, *msg_date, *folder_id;
	guint msgnum;
	GtkListStore *store = ftsearch_window.store;
	GtkTreeIter iter;
	GString *msg;
	gchar *str;
	gint nmsg = 0;
	gboolean ret = FALSE;
	gint error = 0;
	gchar *sep_str = NULL;

	gtk_list_store_clear(store);
	update_text("");
	gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label), "");
	query_info_free(query_info);
	query_info = NULL;

	if (qinfo->body || qinfo->from || qinfo->to || qinfo->subject) {
		have_cond = TRUE;
	}

	if (!have_cond) {
		gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label),
				   _("Search keyword is missing."));
		return -1;
	}

	msg = g_string_new("SEARCHMSG\n");

	if (qinfo->body) {
		g_string_append_printf(msg, "BODY:%s\n", qinfo->body);
	}
	if (qinfo->from) {
		g_string_append_printf(msg, "FROM:%s\n", qinfo->from);
	}
	if (qinfo->to) {
		g_string_append_printf(msg, "TO:%s\n", qinfo->to);
	}
	if (qinfo->subject) {
		g_string_append_printf(msg, "SUBJECT:%s\n", qinfo->subject);
	}
	if (qinfo->date_gt) {
		g_string_append_printf(msg, "DATE-GT:%s\n", qinfo->date_gt);
	}
	if (qinfo->date_lt) {
		g_string_append_printf(msg, "DATE-LT:%s\n", qinfo->date_lt);
	}
	if (qinfo->folder) {
		if (qinfo->folder_with_sub)
			g_string_append_printf(msg, "FOLDER-WITH-SUB:%s\n", qinfo->folder);
		else
			g_string_append_printf(msg, "FOLDER:%s\n", qinfo->folder);
	}
	if (qinfo->has_attach) {
		g_string_append_printf(msg, "HAVE-ATTACH:1\n");
	}
	if (qinfo->res_limit > 0) {
		g_string_append_printf(msg, "RES-LIMIT:%d\n", qinfo->res_limit);
	}

	g_string_append(msg, "END");

	ext_send_msg("%s", msg->str);
	while ((ret = ext_recv_line(&str))) {
		gchar **row;

		g_strchomp(str);
		if (g_ascii_strcasecmp(str, "OK") == 0) {
			g_free(str);
			break;
		} else if (g_ascii_strncasecmp(str, "ERROR", 5) == 0) {
			error = atoi(str + 5);
			g_warning("ERROR received (%d)", error);
			if (error == 0)
				error = -1;
			g_free(str);
			break;
		}

		row = g_strsplit(str, "\t", -1);
		if (g_strv_length(row) < 7) {
			g_warning("invalid row: %s", str);
			g_strfreev(row);
			g_free(str);
			continue;
		}
		sid = strtoul(row[0], NULL, 10);
		from = row[1];
		to = row[2];
		subject = row[3];
		msg_date = row[4];
		folder_id = row[5];
		msgnum = strtoul(row[6], NULL, 10);

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				   COL_SUBJECT, subject,
				   COL_FROM, from,
				   COL_DATE, msg_date,
				   COL_FOLDER, folder_id,
				   COL_MSGNUM, msgnum,
				   COL_MSGSID, sid,
				   -1);

		g_strfreev(row);
		g_free(str);

		nmsg++;
	}

	if (!ret)
		error = -1;
	if (error != 0)
		str = g_strdup_printf(_("Error occurred (%d: %s)"),
				      error, errcode_to_str(error));
	else
		str = g_strdup_printf(_("%d messages found."), nmsg);
	g_print("%s\n\n", str);
	gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label), str);
	g_free(str);

	if (error != 0)
		return -1;

	if (qinfo->body) {
		ext_send_msg("WAKACHI\n%s", qinfo->body);
		if (!ext_recv_line(&sep_str))
			return -1;
		g_strchomp(sep_str);
	}

	query_info = query_info_new(qinfo->body, qinfo->from, qinfo->to,
				    qinfo->subject);
	query_info->sep_body = sep_str;
	g_print("search done\n");

	return 0;
}

static gchar *get_body_text(gulong sid)
{
	gchar *str;
	guint bytes;

	ext_send_msg("GETBODY %lu", sid);
	if (!ext_recv_line(&str))
		return NULL;
	g_strchomp(str);
	if (g_ascii_strncasecmp(str, "ERROR", 5) == 0) {
		gint error;

		error = atoi(str + 5);
		g_warning("ERROR received (%d)", error);
		g_free(str);
		return NULL;
	}
	bytes = strtoul(str, NULL, 10);
	g_free(str);

	if (!ext_recv_bytes(bytes, &str))
		return NULL;

	return str;
}

static gboolean gtkut_text_buffer_match_string(GtkTextBuffer *textbuf,
					       const GtkTextIter *iter,
					       gunichar *wcs, gint len,
					       gboolean case_sens)
{
	GtkTextIter start_iter, end_iter;
	gchar *utf8str, *p;
	gint match_count;

	start_iter = end_iter = *iter;
	gtk_text_iter_forward_chars(&end_iter, len);

	utf8str = gtk_text_buffer_get_text(textbuf, &start_iter, &end_iter,
					   FALSE);
	if (!utf8str) return FALSE;

	if ((gint)g_utf8_strlen(utf8str, -1) != len) {
		g_free(utf8str);
		return FALSE;
	}

	for (p = utf8str, match_count = 0; *p != '\0' && match_count < len;
	     p = g_utf8_next_char(p), match_count++) {
		gunichar wc;

		wc = g_utf8_get_char(p);

		if (case_sens) {
			if (wc != wcs[match_count])
				break;
		} else {
			if (g_unichar_tolower(wc) !=
			    g_unichar_tolower(wcs[match_count]))
				break;
		}
	}

	g_free(utf8str);

	if (match_count == len)
		return TRUE;
	else
		return FALSE;
}

static gboolean gtkut_text_buffer_find(GtkTextBuffer *buffer,
				       const GtkTextIter *iter,
				       const gchar *str, gboolean case_sens,
				       GtkTextIter *match_pos)
{
	gunichar *wcs;
	gint len;
	glong items_read = 0, items_written = 0;
	GError *error = NULL;
	GtkTextIter iter_;
	gboolean found = FALSE;

	wcs = g_utf8_to_ucs4(str, -1, &items_read, &items_written, &error);
	if (error != NULL) {
		g_warning("An error occured while converting a string from UTF-8 to UCS-4: %s\n", error->message);
		g_error_free(error);
	}
	if (!wcs || items_written <= 0) return FALSE;
	len = (gint)items_written;

	iter_ = *iter;
	do {
		found = gtkut_text_buffer_match_string
			(buffer, &iter_, wcs, len, case_sens);
		if (found) {
			*match_pos = iter_;
			break;
		}
	} while (gtk_text_iter_forward_char(&iter_));

	g_free(wcs);

	return found;
}

static void update_text(const gchar *text)
{
	GtkTextBuffer *buffer;
	GtkTextIter iter, pos;
	GtkTextMark *mark;
	gchar **keywords;
	gint i, len;

	if (!text)
		text = "";

	buffer = gtk_text_view_get_buffer
		(GTK_TEXT_VIEW(ftsearch_window.textview));
	gtk_text_buffer_set_text(buffer, text, -1);
	gtk_text_buffer_get_start_iter(buffer, &iter);
	gtk_text_buffer_place_cursor(buffer, &iter);
	mark = gtk_text_buffer_get_insert(buffer);
	gtk_text_view_scroll_mark_onscreen
		(GTK_TEXT_VIEW(ftsearch_window.textview), mark);

	if (!query_info || !query_info->sep_body)
		return;

	keywords = g_strsplit(query_info->sep_body, " ", 0);

	for (i = 0; keywords[i] != NULL; i++) {
		gtk_text_buffer_get_start_iter(buffer, &iter);
		len = g_utf8_strlen(keywords[i], -1);
		while (gtkut_text_buffer_find(buffer, &iter, keywords[i], FALSE,
					      &pos)) {
			GtkTextIter end = pos;

			gtk_text_iter_forward_chars(&end, len);
			gtk_text_buffer_apply_tag(buffer, ftsearch_window.tag,
						  &pos, &end);
			iter = end;
		}
	}

	g_strfreev(keywords);
}

static QueryInfo *query_info_new(const gchar *body, const gchar *from,
				 const gchar *to, const gchar *subject)
{
	QueryInfo *qinfo;

	qinfo = g_new0(QueryInfo, 1);
	qinfo->body = g_strdup(body);
	qinfo->from = g_strdup(from);
	qinfo->to = g_strdup(to);
	qinfo->subject = g_strdup(subject);

	return qinfo;
}

static void query_info_free(QueryInfo *qinfo)
{
	if (!qinfo)
		return;

	g_free(qinfo->body);
	g_free(qinfo->from);
	g_free(qinfo->to);
	g_free(qinfo->subject);
	g_free(qinfo->date_gt);
	g_free(qinfo->date_lt);
	g_free(qinfo->folder);
	g_free(qinfo->sep_body);
	g_free(qinfo);
}

static const gchar *errcode_to_str(gint error)
{
	switch (error) {
	case 0:
		return _("Success");
	case SP_ERROR_DBCONN: 
		return _("DB connection failed");
	case SP_ERROR_DBQUERY:
		return _("DB query failed");
	case SP_ERROR_INITDB:
		return _("DB initialization failed");
	case SP_ERROR_CREATEDB:
		return _("DB creation failed");
	case SP_ERROR_STARTDB:
		return _("DB start failed");
	case SP_ERROR_DB:
		return _("DB error");
	case SP_ERROR_LICENSE_NG:
		return _("License auth failed");
	case SP_ERROR_ARG:
		return _("Invalid argument");
	default:
		return _("Internal error");
	}
}

static void syl_searcher_create(void)
{
	GtkWidget *window;
	GtkWidget *vbox_top;
	GtkWidget *vbox;

	GdkPixbuf *pixbuf;
	GtkWidget *icon;

	GtkActionGroup *group;
	GtkUIManager *ui;
	GtkWidget *menubar;

	GtkWidget *table;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *body_entry;
	GtkWidget *from_entry;
	GtkWidget *to_entry;
	GtkWidget *subject_entry;
	GtkWidget *dategt_entry;
	GtkWidget *datelt_entry;
	GtkTooltips *tip;
	GtkWidget *folder_entry;
	GtkWidget *folder_btn;
	GtkWidget *folder_with_sub_chkbtn;
	GtkWidget *has_attach_chkbtn;

	GtkWidget *vpaned;

	GtkWidget *scrolledwin;

	GtkWidget *treeview;
	GtkListStore *store;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkTreeSelection *selection;

	GtkWidget *textview;
	GtkTextBuffer *buffer;
	GtkTextTag *tag;

	GtkWidget *status_label;

	GtkWidget *hbbox;
	GtkWidget *search_btn;
	GtkWidget *clear_btn;
	GtkWidget *close_btn;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW (window),
			     _("Full-text search messages"));
	gtk_widget_set_size_request(window, 620, -1);
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, TRUE, TRUE);
	g_signal_connect(G_OBJECT(window), "delete_event",
			 G_CALLBACK(search_deleted), NULL);
	syl_plugin_manage_window_signals_connect(GTK_WINDOW(window));

	pixbuf = gdk_pixbuf_new_from_inline(sizeof(sylpheed_pro_icon_32),
					    sylpheed_pro_icon_32, 
					    FALSE, NULL);
	gtk_window_set_icon(GTK_WINDOW(window), pixbuf);

	vbox_top = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(window), vbox_top);

	group = gtk_action_group_new("main");
	gtk_action_group_set_translation_domain(group, GETTEXT_PACKAGE);
	gtk_action_group_add_actions(group, action_entries,
				     sizeof(action_entries) /
				     sizeof(action_entries[0]), NULL);

	ui = gtk_ui_manager_new();
	gtk_ui_manager_insert_action_group(ui, group, 0);
	gtk_ui_manager_add_ui_from_string(ui, ui_def, -1, NULL);
	menubar = gtk_ui_manager_get_widget(ui, "/MainMenu");
	gtk_box_pack_start(GTK_BOX(vbox_top), menubar, FALSE, FALSE, 0);

	vbox = gtk_vbox_new(FALSE, 6);
	gtk_box_pack_start(GTK_BOX(vbox_top), vbox, TRUE, TRUE, 0);
	gtk_container_set_border_width(GTK_CONTAINER (vbox), 8);

	table = gtk_table_new(5, 2, FALSE);
	gtk_table_set_col_spacings(GTK_TABLE(table), 4);
	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("Body:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);

	body_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), body_entry, 1, 2, 0, 1);
	g_signal_connect(G_OBJECT(body_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tip, body_entry,
			     _("Enter words or sentence included in body text\n"
			       "(Sentence is automatically divided into words)"),
			     NULL);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("From:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);

	from_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), from_entry, 1, 2, 1, 2);
	g_signal_connect(G_OBJECT(from_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tip, from_entry,
			     _("Specify part of name or address of sender (From)"),
			     NULL);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("To:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);

	to_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), to_entry, 1, 2, 2, 3);
	g_signal_connect(G_OBJECT(to_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	gtk_tooltips_set_tip(tip, to_entry,
			     _("Specify part of name or address of recipient (To or Cc)"),
			     NULL);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("Subject:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 3, 4, GTK_FILL, 0, 0, 0);

	subject_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), subject_entry, 1, 2, 3, 4);
	g_signal_connect(G_OBJECT(subject_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	gtk_tooltips_set_tip(tip, subject_entry, _("Specify part of subject"),
			     NULL);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("Folder:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 4, 5, GTK_FILL, 0, 0, 0);

	hbox = gtk_hbox_new(FALSE, 8);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 2, 4, 5);

	folder_entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(hbox), folder_entry, TRUE, TRUE, 0);
	g_signal_connect(G_OBJECT(folder_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tip, folder_entry, _("Specify search target folder"),
			     NULL);

	folder_btn = gtk_button_new_with_label("...");
	gtk_box_pack_start(GTK_BOX(hbox), folder_btn, FALSE, FALSE, 0);

	folder_with_sub_chkbtn = gtk_check_button_new_with_label(_("Search subfolders"));
	gtk_box_pack_start(GTK_BOX(hbox), folder_with_sub_chkbtn, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 8);
	label = gtk_label_new(_("Date newer than:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	dategt_entry = gtk_entry_new();
	gtk_widget_set_size_request(dategt_entry, 130, -1);
	gtk_box_pack_start(GTK_BOX(hbox), dategt_entry, FALSE, FALSE, 0);
	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tip, dategt_entry, _("Date format:\n"
						  "  2007/1/15\n"
						  "  2007/02/01 12:34"),
			     NULL);

	label = gtk_label_new(_("older than:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	datelt_entry = gtk_entry_new();
	gtk_widget_set_size_request(datelt_entry, 130, -1);
	gtk_box_pack_start(GTK_BOX(hbox), datelt_entry, FALSE, FALSE, 0);
	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(tip, datelt_entry, _("Date format:\n"
						  "  2007/1/15\n"
						  "  2007/02/01 12:34"),
			     NULL);

	has_attach_chkbtn = gtk_check_button_new_with_label(_("Attachments"));
	gtk_box_pack_start(GTK_BOX(hbox), has_attach_chkbtn, FALSE, FALSE, 0);

	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 5, 6, GTK_FILL, 0, 0, 0);

	g_signal_connect(G_OBJECT(dategt_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);
	g_signal_connect(G_OBJECT(datelt_entry), "activate",
			 G_CALLBACK(search_entry_activated), NULL);

	vpaned = gtk_vpaned_new();
	gtk_box_pack_start(GTK_BOX(vbox), vpaned, TRUE, TRUE, 0);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_paned_add1(GTK_PANED(vpaned), scrolledwin);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwin),
					    GTK_SHADOW_IN);
	gtk_widget_set_size_request(scrolledwin, -1, 200);

	store = gtk_list_store_new(N_COLS,
				   G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
				   G_TYPE_STRING, G_TYPE_UINT, G_TYPE_ULONG);
	treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
	g_object_unref(store);
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeview), TRUE);
	g_signal_connect(G_OBJECT(treeview), "row-activated",
			 G_CALLBACK(row_activated), NULL);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	g_signal_connect(G_OBJECT(selection), "changed",
			 G_CALLBACK(selection_changed), NULL);

	gtk_container_add(GTK_CONTAINER(scrolledwin), treeview);

#define APPEND_COLUMN(label, col, width)                                \
{                                                                       \
        renderer = gtk_cell_renderer_text_new();                        \
        column = gtk_tree_view_column_new_with_attributes               \
                (label, renderer, "text", col, NULL);                   \
        gtk_tree_view_column_set_resizable(column, TRUE);               \
        if (width) {                                                    \
                gtk_tree_view_column_set_sizing                         \
                        (column, GTK_TREE_VIEW_COLUMN_FIXED);           \
                gtk_tree_view_column_set_fixed_width(column, width);    \
        }                                                               \
        gtk_tree_view_column_set_sort_column_id(column, col);           \
        gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);   \
}

        APPEND_COLUMN(_("Subject"), COL_SUBJECT, 240);
        APPEND_COLUMN(_("From"), COL_FROM, 180);
        APPEND_COLUMN(_("Date"), COL_DATE, 0);
        APPEND_COLUMN(_("Folder"), COL_FOLDER, 0);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_paned_add2(GTK_PANED(vpaned), scrolledwin);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwin),
					    GTK_SHADOW_IN);

	textview = gtk_text_view_new();
	gtk_widget_set_size_request(textview, -1, 200);
	gtk_container_add(GTK_CONTAINER(scrolledwin), textview);
        gtk_text_view_set_editable(GTK_TEXT_VIEW(textview), FALSE);
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(textview), GTK_WRAP_WORD);
	gtk_text_view_set_left_margin(GTK_TEXT_VIEW(textview), 6);
	gtk_text_view_set_right_margin(GTK_TEXT_VIEW(textview), 6);

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));
	tag = gtk_text_buffer_create_tag(buffer, "match-string",
					 "background", "blue",
					 "foreground", "white",
					  NULL);		 

	hbox = gtk_hbox_new(FALSE, 8);
	gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	icon = gtk_image_new_from_pixbuf(pixbuf);
	g_object_unref(pixbuf);
	gtk_box_pack_start(GTK_BOX(hbox), icon, FALSE, FALSE, 0);

	status_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), status_label, FALSE, FALSE, 0);

	hbbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbbox), GTK_BUTTONBOX_END);
	gtk_box_set_spacing(GTK_BOX(hbbox), 6);

	search_btn = gtk_button_new_from_stock(GTK_STOCK_FIND);
	GTK_WIDGET_SET_FLAGS(search_btn, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(hbbox), search_btn, FALSE, FALSE, 0);

	clear_btn = gtk_button_new_from_stock(GTK_STOCK_CLEAR);
	GTK_WIDGET_SET_FLAGS(clear_btn, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(hbbox), clear_btn, FALSE, FALSE, 0);

	close_btn = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
	GTK_WIDGET_SET_FLAGS(close_btn, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(hbbox), close_btn, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(folder_btn), "clicked",
			 G_CALLBACK(folder_btn_clicked), NULL);
	g_signal_connect(G_OBJECT(search_btn), "clicked",
			 G_CALLBACK(search_clicked), NULL);
	g_signal_connect(G_OBJECT(clear_btn), "clicked",
			 G_CALLBACK(clear_clicked), NULL);
	g_signal_connect(G_OBJECT(close_btn), "clicked",
			 G_CALLBACK(close_clicked), NULL);

	gtk_box_pack_end(GTK_BOX(hbox), hbbox, FALSE, FALSE, 0);

	gtk_widget_show_all(window);

	ftsearch_window.window = window;
	ftsearch_window.body_entry = body_entry;
	ftsearch_window.from_entry = from_entry;
	ftsearch_window.to_entry = to_entry;
	ftsearch_window.subject_entry = subject_entry;
	ftsearch_window.dategt_entry = dategt_entry;
	ftsearch_window.datelt_entry = datelt_entry;
	ftsearch_window.folder_entry = folder_entry;
	ftsearch_window.folder_btn = folder_btn;
	ftsearch_window.folder_with_sub_chkbtn = folder_with_sub_chkbtn;
	ftsearch_window.has_attach_chkbtn = has_attach_chkbtn;
	ftsearch_window.treeview = treeview;
	ftsearch_window.store = store;
	ftsearch_window.textview = textview;
	ftsearch_window.tag = tag;
	ftsearch_window.status_label = status_label;
	ftsearch_window.search_btn = search_btn;
	ftsearch_window.clear_btn = clear_btn;
	ftsearch_window.close_btn = close_btn;
}

void alert_dialog(const gchar *title, const gchar *msg, GtkWindow *parent)
{
	GtkWidget *dialog;

	if (parent)
		gtk_window_present(parent);

	dialog = gtk_message_dialog_new(parent,
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK,
					"%s", title);
#if GTK_CHECK_VERSION(2, 6, 0)
	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
						 "%s", msg);
#endif
	gtk_widget_show(dialog);
	gtk_window_set_transient_for(GTK_WINDOW(dialog), parent);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

static gboolean read_config_done = FALSE;

gint searcher_init(void)
{
	static gboolean init_done = FALSE;

	g_print("searcher_init\n");

	if (init_done) {
		if (ftsearch_window.window)
			gtk_window_present(GTK_WINDOW(ftsearch_window.window));
		return 0;
	}
	init_done = TRUE;

	syl_searcher_create();

	if (searcher_read_config(FALSE) < 0) {
		if (searcher_config(&config, GTK_WINDOW(ftsearch_window.window)))
			read_config_done = TRUE;
	}

	return 0;
}

gint searcher_read_config(gboolean all_check)
{
	if (!read_config_done && read_config(&config) < 0) {
		g_print("searcher_read_config: config not exist. Using default config\n");
		if (!config.dbname)
			config.dbname = g_strdup("sylph");
		config.port = DEFAULT_PORT;
		if (!config.user)
			config.user = g_strdup(DEFAULT_USER);
		config.auto_update = TRUE;
		if (!config.exclude_list) {
			config.exclude_list = g_strdup("junk,spam,trash");
			update_exclude_list(&config);
		}
		config.enable_searcher = TRUE;
		return -1;
	}

	if (config.enable_searcher || all_check) {
		if (config.dbpath && *config.dbpath &&
		    !is_dir_exist(config.dbpath)) {
			g_warning("Database directory %s does not exist.", config.dbpath);
			read_config_done = FALSE;
			return -1;
		} else if (!config.dbpath || *config.dbpath == '\0') {
			g_warning("Database directory not set.");
			read_config_done = FALSE;
			return -1;
		} else {
			gchar *key;

			key = read_license_key();
			if (!key || *key == '\0') {
				g_warning("License key not exist.");
				g_free(key);
				read_config_done = FALSE;
				return -1;
			}
			g_free(key);
		}
	}

	read_config_done = TRUE;
	return 0;
}

void searcher_settings(void)
{
	GtkWindow *window = NULL;

	if (searcher_read_config(TRUE) < 0) {
		searcher_setup();
		return;
	}

	if (ftsearch_window.window &&
	    GTK_WIDGET_VISIBLE(ftsearch_window.window))
		window = GTK_WINDOW(ftsearch_window.window);

	if (searcher_config(&config, window))
		read_config_done = TRUE;
}

AppConfig *searcher_get_config(void)
{
	return &config;
}

void searcher_clear(void)
{
#if GTK_CHECK_VERSION(2, 6, 0)
	gtk_tree_sortable_set_sort_column_id
		(GTK_TREE_SORTABLE(ftsearch_window.store),
		 GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, GTK_SORT_ASCENDING);
#endif
	gtk_list_store_clear(ftsearch_window.store);

	update_text("");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.body_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.from_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.to_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.subject_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.dategt_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.datelt_entry), "");
	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.folder_entry), "");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ftsearch_window.folder_with_sub_chkbtn), FALSE);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ftsearch_window.has_attach_chkbtn), FALSE);
	gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label), "");
	query_info_free(query_info);
	query_info = NULL;
}

void searcher_set_folder(const gchar *folder_id)
{
	if (folder_id)
		gtk_entry_set_text(GTK_ENTRY(ftsearch_window.folder_entry),
				   folder_id);
} 

static struct SetupDialog
{
	GtkWidget *dialog;
	GtkWidget *vbox_p1, *vbox_p2, *vbox_p3, *vbox_p4, *vbox_pe;
	GtkWidget *prev_btn, *next_btn, *cancel_btn;
	GtkWidget *license_entry;
	GtkWidget *license_radiobtn;
	GtkWidget *dbpath_entry;
	GtkWidget *license_label;
	GtkWidget *dbpath_label;
	GtkWidget *message_label;
	gint page;
	gboolean enable_trial;
	gboolean not_again;
	gboolean finished;
} setup_dialog;

static gint searcher_init_database(const gchar *dbpath)
{
	gchar buf[64];
	gchar *pgfile;
	gboolean db_already_exist = FALSE;
	gboolean create_db = TRUE;

	g_free(config.dbname);
	config.dbname = g_strdup("sylph");
	config.port = DEFAULT_PORT;
	g_free(config.user);
	config.user = g_strdup(DEFAULT_USER);
	g_free(config.dbpath);
	config.dbpath = g_strdup(dbpath);
	config.auto_update = TRUE;
	if (!config.exclude_list) {
		config.exclude_list = g_strdup("junk,spam,trash");
		update_exclude_list(&config);
	}
	config.enable_searcher = TRUE;
	write_config(&config);
	ext_send_msg("RELOAD");
	if (!ext_recv_msg(buf, sizeof(buf)))
		return -1;
	if (g_ascii_strcasecmp(buf, "OK") != 0)
		return -1;

	pgfile = g_strconcat(dbpath, G_DIR_SEPARATOR_S, "PG_VERSION", NULL);
	if (is_dir_exist(dbpath) && is_file_exist(pgfile)) {
		g_print("%s already exist. skipping initdb\n", pgfile);
		g_free(pgfile);
		db_already_exist = TRUE;
	} else {
		g_free(pgfile);
		ext_send_msg("INITDB\n%s", dbpath);
		if (!ext_recv_msg(buf, sizeof(buf)))
			return -1;
		if (g_ascii_strcasecmp(buf, "OK") != 0)
			return -1;
	}

	gtk_label_set_text(GTK_LABEL(setup_dialog.message_label),
			   _("Starting database..."));
	ui_update();

	ext_send_msg("STARTDB");
	if (!ext_recv_msg(buf, sizeof(buf)))
		return -1;
	if (g_ascii_strcasecmp(buf, "OK") != 0)
		return -1;

	if (db_already_exist) {
		GtkWidget *dialog;
		gint response;

		dialog = gtk_message_dialog_new(GTK_WINDOW(setup_dialog.dialog),
						GTK_DIALOG_DESTROY_WITH_PARENT,
						GTK_MESSAGE_QUESTION,
						GTK_BUTTONS_YES_NO,
						"%s", _("Reset database"));
#if GTK_CHECK_VERSION(2, 6, 0)
		gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), "%s", _("Database already exists. Delete current database and re-create new one?"));
#endif
		gtk_widget_show(dialog);
		gtk_window_set_transient_for(GTK_WINDOW(dialog),
					     GTK_WINDOW(setup_dialog.dialog));
		response = gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);

		if (response != GTK_RESPONSE_YES)
			create_db = FALSE;
	}

	if (create_db) {
		gtk_label_set_text(GTK_LABEL(setup_dialog.message_label),
				   _("Creating database..."));
		ui_update();

		ext_send_msg("CREATEDB");
		if (!ext_recv_msg(buf, sizeof(buf)))
			return -1;
		if (g_ascii_strcasecmp(buf, "OK") != 0)
			return -1;

		ext_send_msg("CHECKDB");
		if (!ext_recv_msg(buf, sizeof(buf)))
			return -1;
		if (g_ascii_strcasecmp(buf, "OK") != 0)
			return -1;
	} else {
		g_print("skipping createdb\n");
	}

	return 0;
}

static void button_toggled(GtkToggleButton *button, GtkWidget *widget)
{
	gboolean is_active;

	is_active = gtk_toggle_button_get_active(button);
	gtk_widget_set_sensitive(widget, is_active);
}

static gint confirm_cancel(gboolean *not_again)
{
	GtkWidget *dialog;
	GtkWidget *checkbtn;
	gint response;

	dialog = gtk_message_dialog_new(GTK_WINDOW(setup_dialog.dialog),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_WARNING,
					GTK_BUTTONS_YES_NO,
					_("Cancel Sylpheed Pro Setup"));
#if GTK_CHECK_VERSION(2, 6, 0)
	gtk_message_dialog_format_secondary_text
		(GTK_MESSAGE_DIALOG(dialog), _("Do you cancel the setup of Sylpheed Pro full-text search?"));
#endif
	checkbtn = gtk_check_button_new_with_label(_("Do not ask again"));
	gtk_widget_show(checkbtn);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), checkbtn,
			   FALSE, FALSE, 0);

	gtk_widget_show(dialog);
	gtk_window_set_transient_for(GTK_WINDOW(dialog),
				     GTK_WINDOW(setup_dialog.dialog));
	response = gtk_dialog_run(GTK_DIALOG(dialog));
	*not_again = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbtn));
	gtk_widget_destroy(dialog);

	g_print("confirm_cancel: response: %d, not_again: %d\n", response == GTK_RESPONSE_YES, *not_again);

	return response;
}

static void setup_response_cb(GtkDialog *dialog, gint response_id,
			      gpointer data)
{
	gchar buf[64] = "";
	const gchar *key;

	if (response_id == GTK_RESPONSE_CANCEL ||
	    response_id == GTK_RESPONSE_DELETE_EVENT) {
		gint resp2;
		gboolean not_again;

		if (setup_dialog.page <= 3) {
			resp2 = confirm_cancel(&not_again);
			if (resp2 == GTK_RESPONSE_YES) {
				setup_dialog.not_again = not_again;
				gtk_widget_destroy(GTK_WIDGET(dialog));
				setup_dialog.dialog = NULL;
			}
		} else if (setup_dialog.page == 4) {
			return;
		} else {
			gtk_widget_destroy(GTK_WIDGET(dialog));
			setup_dialog.dialog = NULL;
		}
	} else if (response_id == GTK_RESPONSE_ACCEPT) {
		switch (setup_dialog.page) {
		case 0:
			gtk_widget_hide(setup_dialog.vbox_p1);
			gtk_widget_show(setup_dialog.vbox_p2);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, TRUE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			break;
		case 1:
			if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(setup_dialog.license_radiobtn))) {
				setup_dialog.enable_trial = FALSE;
				key = gtk_entry_get_text(GTK_ENTRY(setup_dialog.license_entry));
				if (check_license_key("user", key) != 0) {
					g_warning("license key check failed: %s", key);
					alert_dialog(_("Error"), _("Invalid license key."), GTK_WINDOW(dialog));
					return;
				}
			} else {
				g_print("enable trial version\n");
				setup_dialog.enable_trial = TRUE;
			}
			gtk_widget_hide(setup_dialog.vbox_p2);
			gtk_widget_show(setup_dialog.vbox_p3);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, TRUE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			break;
		case 2:
			gtk_widget_hide(setup_dialog.vbox_p3);
			gtk_widget_show(setup_dialog.vbox_p4);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, TRUE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			if (setup_dialog.enable_trial)
				gtk_label_set_text(GTK_LABEL(setup_dialog.license_label), _("Trial"));
			else
				gtk_label_set_text(GTK_LABEL(setup_dialog.license_label), gtk_entry_get_text(GTK_ENTRY(setup_dialog.license_entry)));
			gtk_label_set_text(GTK_LABEL(setup_dialog.dbpath_label), gtk_entry_get_text(GTK_ENTRY(setup_dialog.dbpath_entry)));
			break;
		case 3:
			gtk_widget_hide(setup_dialog.vbox_p4);
			gtk_widget_show(setup_dialog.vbox_pe);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, FALSE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, FALSE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_CANCEL, FALSE);
			setup_dialog.page++;
			ui_update();

			key = NULL;
			if (setup_dialog.enable_trial) {
				if (get_trial_key(buf, sizeof(buf)) < 0) {
					gtk_label_set_text(GTK_LABEL(setup_dialog.message_label), _("Couldn't setup trial key"));
				} else {
					key = buf;
					if (check_license_key("user", key) < 0) {
						g_warning("license key check failed: %s", key);
						key = NULL;
					}
				}
			} else {
				key = gtk_entry_get_text(GTK_ENTRY(setup_dialog.license_entry));
			}

			if (key && write_license_key(key) != 0) {
				gtk_label_set_text(GTK_LABEL(setup_dialog.message_label),
						   _("Failed to write license key to file."));
			}

			if (key) {
				if (searcher_init_database(gtk_entry_get_text(GTK_ENTRY(setup_dialog.dbpath_entry))) == 0) {
					gtk_label_set_text(GTK_LABEL(setup_dialog.message_label),
							   _("Setup completed.\n\nYou can start full-text search by 'Edit - Full-text search' in the main menu,\nor 'Full-text search this folder' in the context menu of folder tree.\n\nPlease update search index by 'Update search index' in the context menu of\nfolder tree before starting full-text search."));
					setup_dialog.finished = TRUE;
				} else {
					gtk_label_set_text(GTK_LABEL(setup_dialog.message_label),
							   _("Initialization of database failed."));
				}
			}

			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_CANCEL, TRUE);
			gtk_button_set_label
				(GTK_BUTTON(setup_dialog.cancel_btn),
				 GTK_STOCK_CLOSE);
			break;
		default:
			break;
		}
		setup_dialog.page++;
	} else if (response_id == GTK_RESPONSE_REJECT) {
		switch (setup_dialog.page) {
		case 1:
			gtk_widget_show(setup_dialog.vbox_p1);
			gtk_widget_hide(setup_dialog.vbox_p2);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, FALSE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			break;
		case 2:
			gtk_widget_show(setup_dialog.vbox_p2);
			gtk_widget_hide(setup_dialog.vbox_p3);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, TRUE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			break;
		case 3:
			gtk_widget_show(setup_dialog.vbox_p3);
			gtk_widget_hide(setup_dialog.vbox_p4);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_REJECT, TRUE);
			gtk_dialog_set_response_sensitive
				(dialog, GTK_RESPONSE_ACCEPT, TRUE);
			break;
		default:
			break;
		}
		setup_dialog.page--;
	}
}

gint searcher_setup(void)
{
	GtkWidget *dialog;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *entry;
	//GtkWidget *spc;
	GtkWidget *vbox2;
	GtkWidget *vbox3;
	GtkWidget *radiobtn;
	GdkPixbuf *pixbuf;
	GtkWidget *logo;
	gchar *str;

	if (setup_dialog.dialog)
		return -1;

	if (start_plugin_system() < 0) {
		return -1;
	}

	dialog = gtk_dialog_new_with_buttons(_("Sylpheed Pro Setup"), NULL, GTK_DIALOG_MODAL, NULL);
	setup_dialog.prev_btn = gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_GO_BACK, GTK_RESPONSE_REJECT);
	setup_dialog.next_btn = gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_GO_FORWARD, GTK_RESPONSE_ACCEPT);
	setup_dialog.cancel_btn = gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
	gtk_widget_set_size_request(dialog, 540, 320);
	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), TRUE);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
	gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
					  GTK_RESPONSE_REJECT, FALSE);

	vbox = gtk_vbox_new(FALSE, 8);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
			   TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	label = gtk_label_new("");
	str = g_strdup_printf("<span size=\"large\" weight=\"bold\">%s</span>",
			      _("Sylpheed Pro Setup"));
	gtk_label_set_markup(GTK_LABEL(label), str);
	g_free(str);
	//gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	pixbuf = gdk_pixbuf_new_from_inline(sizeof(sylpheed_pro_logo_42),
					    sylpheed_pro_logo_42,
					    FALSE, NULL);
	logo = gtk_image_new_from_pixbuf(pixbuf);
	g_object_unref(pixbuf);
	gtk_box_pack_end(GTK_BOX(hbox), logo, FALSE, FALSE, 0);

#if 0
	spc = gtk_vbox_new(FALSE, 0);
	gtk_widget_set_size_request(spc, -1, 8);
	gtk_box_pack_start(GTK_BOX(vbox), spc, FALSE, FALSE, 0);
#endif

	/* First page */
	vbox2 = gtk_vbox_new(FALSE, 20);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Sylpheed Pro full-text search plug-in will enable you to search the body text,\nincluding attached documents such as Microsoft Office, OpenOffice.org and PDF,\nin a pile of messages you have in no time."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Sylpheed Pro full-text search plug-in utilizes the full-text search engine (tsearch2)\nof PostgreSQL database."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Search index will be automatically updated when you receive,\nmove and delete messages."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("This setup will configure full-text search feature of Sylpheed Pro."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	setup_dialog.vbox_p1 = vbox2;

	/* Page 2 */
	vbox2 = gtk_vbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 0);

	radiobtn = gtk_radio_button_new_with_label
		(NULL, _("Enter the license key you received:"));
	gtk_box_pack_start(GTK_BOX(vbox2), radiobtn, FALSE, FALSE, 0);
	setup_dialog.license_radiobtn = radiobtn;

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
	setup_dialog.license_entry = entry;

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

	radiobtn = gtk_radio_button_new_with_label_from_widget
		(GTK_RADIO_BUTTON(setup_dialog.license_radiobtn),
		 _("Try Sylpheed Pro (14 days available)"));
	gtk_box_pack_start(GTK_BOX(vbox2), radiobtn, FALSE, FALSE, 0);

	//gtk_widget_set_sensitive(hbox, FALSE);
	g_signal_connect(G_OBJECT(setup_dialog.license_radiobtn), "toggled",
			 G_CALLBACK(button_toggled), entry);

	setup_dialog.vbox_p2 = vbox2;

	/* Page 3 */
	vbox2 = gtk_vbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("A database for full-text search index will be created at the following location:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(vbox2), entry, FALSE, FALSE, 0);
	setup_dialog.dbpath_entry = entry;

	vbox3 = gtk_vbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(vbox2), vbox3, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("* The database will become large size, so please select the drive\n  with large free space."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("* It is recommended to exclude the database folder from automatic anti-virus\n  protection to avoid troubles like errors or low performance."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	setup_dialog.vbox_p3 = vbox2;

	/* Page 4 */
	vbox2 = gtk_vbox_new(FALSE, 24);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Sylpheed Pro will be initialized with the following setting:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("License key:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	setup_dialog.license_label = label;

	hbox = gtk_hbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Database location:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	setup_dialog.dbpath_label = label;

	setup_dialog.vbox_p4 = vbox2;

	/* Last page */
	vbox2 = gtk_vbox_new(FALSE, 24);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
	label = gtk_label_new(_("Initializing database..."));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	setup_dialog.message_label = label;

	setup_dialog.vbox_pe = vbox2;

	g_signal_connect(G_OBJECT(dialog), "response",
			 G_CALLBACK(setup_response_cb), NULL);

	syl_plugin_manage_window_set_transient(GTK_WINDOW(dialog));

	gtk_widget_show_all(dialog);
	gtk_widget_hide(setup_dialog.vbox_p2);
	gtk_widget_hide(setup_dialog.vbox_p3);
	gtk_widget_hide(setup_dialog.vbox_p4);
	gtk_widget_hide(setup_dialog.vbox_pe);
	setup_dialog.dialog = dialog;
	setup_dialog.page = 0;
	setup_dialog.finished = FALSE;

	str = read_license_key();
	if (str && *str != '\0') {
		if (check_license_key_check_only("user", str) == 0)
			gtk_entry_set_text(GTK_ENTRY(setup_dialog.license_entry), str);
	}
	g_free(str);

	if (config.dbpath) {
		gtk_entry_set_text(GTK_ENTRY(setup_dialog.dbpath_entry), config.dbpath);
	} else {
		str = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, "searcherdb", NULL);
		gtk_entry_set_text(GTK_ENTRY(setup_dialog.dbpath_entry), str);
		g_free(str);
	}

	syl_plugin_inc_lock();
	while (setup_dialog.dialog)
		gtk_dialog_run(GTK_DIALOG(setup_dialog.dialog));
	syl_plugin_inc_unlock();

	g_print("dialog end\n");
	if (setup_dialog.finished) {
		g_print("setup completed\n");
		return 0;
	} else {
		if (setup_dialog.not_again) {
			g_print("do not show setup again\n");
			config.enable_searcher = FALSE;
			write_config(&config);
		}
		return -1;
	}
}

static void selection_changed(GtkTreeSelection *selection, gpointer data)
{
	GtkTreeModel *model;
	gulong sid = 0;
	gchar *text;
	GtkTreeIter iter;

	if (!gtk_tree_selection_get_selected(selection, &model, &iter)) {
		update_text("");
		return;
	}

	gtk_tree_model_get(model, &iter, COL_MSGSID, &sid, -1);
	if (!sid) {
		update_text("");
		return;
	}

	text = get_body_text(sid);
	update_text(text);
	g_free(text);
}

static void row_activated(GtkTreeView *treeview, GtkTreePath *path,
			  GtkTreeViewColumn *column, gpointer data)
{
	GtkTreeModel *model = GTK_TREE_MODEL(ftsearch_window.store);
	GtkTreeIter iter;
	gchar *folder_id;
	guint msgnum;

	if (gtk_tree_model_get_iter(model, &iter, path) == FALSE)
		return;

	gtk_tree_model_get(model, &iter, COL_FOLDER, &folder_id,
			   COL_MSGNUM, &msgnum, -1);
	if (!folder_id || *folder_id == '\0')
		return;

	g_print("opening '%s/%u' on Sylpheed\n", folder_id, msgnum);
	syl_plugin_open_message(folder_id, msgnum);
}

static void search_entry_activated(GtkWidget *widget, gpointer data)
{
	gtk_button_clicked(GTK_BUTTON(ftsearch_window.search_btn));
}

static gint search_deleted(GtkWidget *widget, GdkEventAny *event,
			   gpointer data)
{
	close_cb();
	return TRUE;
}

static void folder_btn_clicked(GtkButton *button, gpointer data)
{
	FolderItem *item;
	gchar *id;

	item = syl_plugin_folder_sel(NULL, 0, NULL);
	if (!item || item->stype == F_VIRTUAL)
		return;

	id = folder_item_get_identifier(item);
	if (id) {
		gtk_entry_set_text(GTK_ENTRY(ftsearch_window.folder_entry), id);
		g_free(id);
	}
}

static void clear_clicked(GtkButton *button, gpointer data)
{
	searcher_clear();
}

static void search_clicked(GtkButton *button, gpointer data)
{
	QueryInfo qinfo = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, FALSE, 0};
	const gchar *str;

	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.body_entry));
	if (str && *str)
		qinfo.body = (gchar *)str;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.from_entry));
	if (str && *str)
		qinfo.from = (gchar *)str;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.to_entry));
	if (str && *str)
		qinfo.to = (gchar *)str;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.subject_entry));
	if (str && *str)
		qinfo.subject = (gchar *)str;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.dategt_entry));
	if (str && *str)
		qinfo.date_gt = (gchar *)str;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.datelt_entry));
	if (str && *str)
		qinfo.date_lt = (gchar *)str;
	qinfo.res_limit = config.res_limit;
	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.folder_entry));
	if (str && *str)
		qinfo.folder = (gchar *)str;
	if (gtk_toggle_button_get_active
		(GTK_TOGGLE_BUTTON(ftsearch_window.folder_with_sub_chkbtn)))
		qinfo.folder_with_sub = TRUE;
	if (gtk_toggle_button_get_active
		(GTK_TOGGLE_BUTTON(ftsearch_window.has_attach_chkbtn)))
		qinfo.has_attach = TRUE;

	do_query(&qinfo);
}

static void close_clicked(GtkButton *button, gpointer data)
{
	gtk_widget_hide(ftsearch_window.window);
}

static void close_cb(void)
{
	gtk_widget_hide(ftsearch_window.window);
}

static void copy_cb(void)
{
	GtkWidget *widget;
	GtkClipboard *clipboard;
	GtkTextBuffer *buffer;

	widget = gtk_window_get_focus(GTK_WINDOW(ftsearch_window.window));
	if (widget == ftsearch_window.textview) {
		buffer = gtk_text_view_get_buffer
			(GTK_TEXT_VIEW(ftsearch_window.textview));
		clipboard = gtk_widget_get_clipboard(ftsearch_window.textview,
						     GDK_SELECTION_CLIPBOARD);
		gtk_text_buffer_copy_clipboard(buffer, clipboard);
	} else if (GTK_IS_EDITABLE(widget)) {
		gtk_editable_copy_clipboard(GTK_EDITABLE(widget));
	}
}

static void paste_cb(void)
{
	GtkWidget *widget;

	widget = gtk_window_get_focus(GTK_WINDOW(ftsearch_window.window));
	if (GTK_IS_EDITABLE(widget)) {
		gtk_editable_paste_clipboard(GTK_EDITABLE(widget));
	}
}

static void settings_cb(void)
{
	searcher_config_open();
}

static void about_cb(void)
{
	about_dialog();
}
